from typing import List

from matplotlib import patches
from matplotlib import pyplot as plt
from torchvision.transforms.v2 import Resize
import numpy as np
import torch

from sihl.heads import KeypointDetection

from .common import get_images, plot_to_numpy, COLORS


def get_rectangle_patch(box, color):
    xy, width, height = (box[0], box[1]), box[2] - box[0], box[3] - box[1]
    return patches.Rectangle(
        xy, width, height, linewidth=1, edgecolor=color, facecolor="none"
    )


@get_images.register(KeypointDetection)
def _(head, config, input, target, features) -> List[np.ndarray]:
    keypoint_labels = config["keypoints"] if "keypoints" in config else []
    keypoint_links = config["links"] if "links" in config else []

    num_instances, scores, _, logits = head.forward_logits(features)
    batch_size, _, height, width = logits.shape
    heatmaps = logits.reshape(
        (batch_size, head.max_instances, head.num_kpts, height, width)
    )
    heatmaps = heatmaps.flatten(3, 4).softmax(dim=3).reshape(heatmaps.shape)
    keypoint_scores = heatmaps.flatten(3, 4).max(dim=3)[0]
    pred_keypoints = head.heatmap_to_keypoints(features, heatmaps)
    scores, reg_features = head.get_features(features)
    scores = scores.max(dim=2)[0]

    images = (input.permute(0, 2, 3, 1) * 255).to(torch.uint8).to("cpu").numpy()
    visualizations = []
    for batch_idx, image in enumerate(images):
        fig, axes = plt.subplots(1, 3, figsize=(10.24, 5.12), dpi=100)
        for ax in axes:
            ax.set_xticks([])
            ax.set_yticks([])
        axes[0].title.set_text("Input")
        axes[0].axis("off")
        axes[0].imshow(image)

        axes[1].title.set_text("Target")
        axes[1].imshow(image, alpha=0.2)
        if target is not None:
            for instance_idx in range(target["presence"][batch_idx].shape[0]):
                presence = target["presence"][batch_idx][instance_idx].to("cpu")
                keypoints = target["keypoints"][batch_idx][instance_idx].to("cpu")
                if keypoints.numel() > 0:
                    x, y = keypoints[:, 0].to("cpu"), keypoints[:, 1].to("cpu")
                    color = np.array(COLORS[instance_idx % len(COLORS)]) / 255
                    axes[1].scatter(
                        x=[kpt[0] for kpt, pres in zip(keypoints, presence) if pres],
                        y=[kpt[1] for kpt, pres in zip(keypoints, presence) if pres],
                        marker="o",
                        s=3,
                        c=color,
                    )
                    for link in keypoint_links:
                        i = keypoint_labels.index(link[0])
                        j = keypoint_labels.index(link[1])
                        if not presence[i] or not presence[j]:
                            continue
                        line = ([x[i], x[j]], [y[i], y[j]])
                        axes[1].plot(*line, c=color)

        axes[2].title.set_text("Prediction")
        total_hm = heatmaps[batch_idx, : target["presence"][batch_idx].shape[0]]
        total_hm = total_hm.sum(dim=(0, 1)).unsqueeze(0)
        resize = Resize(image.shape[:2], interpolation=0)
        axes[2].imshow(resize(total_hm).squeeze(0).to("cpu"))
        if pred_keypoints is not None:
            for instance_idx in range(target["presence"][batch_idx].shape[0]):
                presence = keypoint_scores[batch_idx, instance_idx].to("cpu") > 0.1
                keypoints = pred_keypoints[batch_idx, instance_idx].to("cpu")
                if keypoints.numel() > 0:
                    axes[2].scatter(
                        x=[kpt[0] for kpt, pres in zip(keypoints, presence) if pres],
                        y=[kpt[1] for kpt, pres in zip(keypoints, presence) if pres],
                        marker="o",
                        s=3,
                        c=color,
                    )
                    x = [kpt[0] for kpt in keypoints]
                    y = [kpt[1] for kpt in keypoints]
                    color = np.array(COLORS[instance_idx % len(COLORS)]) / 255
                    for link in keypoint_links:
                        i = keypoint_labels.index(link[0])
                        j = keypoint_labels.index(link[1])
                        if not (presence[i] and presence[j]):
                            continue
                        line = ([x[i], x[j]], [y[i], y[j]])
                        axes[2].plot(*line, marker="o", markersize=3, c=color)

        fig.tight_layout()
        visualizations.append(plot_to_numpy(fig))
        plt.close()
    return visualizations



from pathlib import Path
from typing import List, Dict, Tuple, Literal, Any
import logging
import sys

from rich.logging import RichHandler
from torch import Tensor
from torch.utils.data import DataLoader
from torchinfo import summary
import kaggle
import lightning
import lightning.pytorch as pl
import torch
import torchvision
import torchvision.transforms.v2 as transforms

sys.path.append("../src")

from sihl import SihlModel, SihlLightningModule, TorchvisionBackbone
from sihl.heads import ObjectDetection
from sihl.layers import FPN

coco_labels = [
    "__background__",
    "person",
    "bicycle",
    "car",
    "motorbike",
    "aeroplane",
    "bus",
    "train",
    "truck",
    "boat",
    "trafficlight",
    "firehydrant",
    "streetsign",
    "stopsign",
    "parkingmeter",
    "bench",
    "bird",
    "cat",
    "dog",
    "horse",
    "sheep",
    "cow",
    "elephant",
    "bear",
    "zebra",
    "giraffe",
    "hat",
    "backpack",
    "umbrella",
    "shoe",
    "eyeglasses",
    "handbag",
    "tie",
    "suitcase",
    "frisbee",
    "skis",
    "snowboard",
    "sportsball",
    "kite",
    "baseballbat",
    "baseballglove",
    "skateboard",
    "surfboard",
    "tennisracket",
    "bottle",
    "plate",
    "wineglass",
    "cup",
    "fork",
    "knife",
    "spoon",
    "bowl",
    "banana",
    "apple",
    "sandwich",
    "orange",
    "broccoli",
    "carrot",
    "hotdog",
    "pizza",
    "donut",
    "cake",
    "chair",
    "sofa",
    "pottedplant",
    "bed",
    "mirror",
    "diningtable",
    "window",
    "desk",
    "toilet",
    "door",
    "tvmonitor",
    "laptop",
    "mouse",
    "remote",
    "keyboard",
    "cellphone",
    "microwave",
    "oven",
    "toaster",
    "sink",
    "refrigerator",
    "blender",
    "book",
    "clock",
    "vase",
    "scissors",
    "teddybear",
    "hairdrier",
    "toothbrush",
    "hairbrush",
]

valid_coco_labels = coco_labels.copy()
for label in [
    "__background__",
    "streetsign",
    "hat",
    "shoe",
    "eyeglasses",
    "plate",
    "mirror",
    "window",
    "desk",
    "door",
    "blender",
    "hairbrush",
]:
    valid_coco_labels.remove(label)


def collate_fn(batch: List[Tuple[Tensor, Dict[Literal["labels", "boxes"], Any]]]):
    return (
        torch.stack([sample[0] for sample in batch]),
        {
            "classes": [
                (
                    torch.tensor(
                        [
                            valid_coco_labels.index(coco_labels[label])
                            for label in sample[1]["labels"]
                        ]
                    )
                    if "labels" in sample[1]
                    else torch.tensor([], dtype=torch.int64)
                )
                for sample in batch
            ],
            "boxes": [
                sample[1]["boxes"].data if "boxes" in sample[1] else torch.tensor([])
                for sample in batch
            ],
        },
    )


class CocoDataModule(pl.LightningDataModule):
    def __init__(self, batch_size: int) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.data_dir = Path(__file__).parent / "data" / "coco_2017"

    def prepare_data(self) -> None:
        if not self.data_dir.exists():
            kaggle.api.authenticate()
            kaggle.api.dataset_download_files(
                "clkmuhammed/microsoft-coco-2017-common-objects-in-context",
                path=self.data_dir,
                unzip=True,
                quiet=False,
            )

    def setup(self, stage: str = "") -> None:
        annot_dir = self.data_dir / "annotations_trainval2017"
        self.trainset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(self.data_dir / "train2017"),
                str(annot_dir / "instances_train2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(800 - 1, max_size=800),
                        transforms.RandomCrop(800, pad_if_needed=True),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            )
        )
        self.valset = torchvision.datasets.wrap_dataset_for_transforms_v2(
            torchvision.datasets.CocoDetection(
                str(self.data_dir / "val2017"),
                str(annot_dir / "instances_val2017.json"),
                transforms=transforms.Compose(
                    [
                        transforms.ToImage(),
                        transforms.Resize(800 - 1, max_size=800),
                        transforms.RandomCrop(800, pad_if_needed=True),
                        transforms.ToDtype(torch.float32, scale=True),
                    ]
                ),
            )
        )

    def train_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.trainset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=4,
            collate_fn=collate_fn,
        )

    def val_dataloader(self) -> DataLoader[tuple[Tensor, Tensor]]:
        return DataLoader(
            self.valset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=4,
            collate_fn=collate_fn,
        )


if __name__ == "__main__":
    FORMAT = "%(message)s"
    logging.basicConfig(
        level="INFO", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
    )
    log = logging.getLogger("rich")
    logger = pl.loggers.TensorBoardLogger(
        save_dir=Path(__file__).parent / "logs", name="object_detection"
    )
    lightning.seed_everything(0)

    trainer = pl.Trainer(
        max_steps=90_000,
        accelerator="gpu",
        logger=logger,
        callbacks=[pl.callbacks.RichProgressBar(leave=True)],
        gradient_clip_val=3,
        precision="16-mixed",
        val_check_interval=0.25,
    )
    with trainer.init_module():
        backbone = TorchvisionBackbone("resnet50", pretrained=True, frozen_levels=1)
        neck = FPN(backbone.out_channels, out_channels=256, bottom_level=3, top_level=7)
        head = ObjectDetection(
            neck.out_channels,
            num_classes=len(valid_coco_labels),
            soft_label_decay_steps=90_000,
        )
        model = SihlLightningModule(
            SihlModel(backbone=backbone, neck=neck, heads=[head]),
            optimizer=torch.optim.SGD,
            optimizer_kwargs={"lr": 1e-2, "weight_decay": 1e-4, "momentum": 0.9},
            scheduler=torch.optim.lr_scheduler.MultiStepLR,
            scheduler_kwargs={
                "milestones": [60_000, 80_000],
                "gamma": 0.1,
                "warmup_batches": 1000,
            },
            data_config={"categories": valid_coco_labels},
        )

    log.info(
        summary(
            model,
            row_settings=("var_names",),
            col_names=("num_params", "trainable"),
            verbose=0,
            depth=4,
        )
    )
    trainer.fit(model, datamodule=CocoDataModule(batch_size=16))
